<?php

namespace Drupal\linktitle\Plugin\Filter;

use Drupal\Component\Utility\Html;
use Drupal\Component\Utility\UrlHelper;
use Drupal\Core\Url;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\filter\FilterProcessResult;
use Drupal\filter\Plugin\FilterBase;
use GuzzleHttp\Client;
use GuzzleHttp\RequestOptions;
use GuzzleHttp\Exception\GuzzleException;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a filter to automatically attach a title attribute to links.
 *
 * @Filter(
 *   id = "link_title",
 *   title = @translation("Link title"),
 *   description = @Translation("Adds a title attribute to links."),
 *   type = Drupal\filter\Plugin\FilterInterface::TYPE_MARKUP_LANGUAGE,
 * )
 */
class LinkTitle extends FilterBase implements ContainerFactoryPluginInterface {

  /**
   * The number of bytes read while seeking the title tag.
   *
   * @var int
   */
  const SEEK_LENGTH = 100;

  /**
   * The http client.
   *
   * @var \GuzzleHttp\Client
   */
  protected $http;

  /**
   * Constructs a \Drupal\linktitle\Plugin\Filter\LinkTitle object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \GuzzleHttp\Client $http_client
   *   The http client.
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    Client $http_client
  ) {

    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->client = $http_client;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('http_client')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function process($text, $langcode) {
    $result = new FilterProcessResult($text);

    $dom_document = Html::load($text);
    foreach ($dom_document->getElementsByTagName('a') as $link) {
      $title = $link->getAttribute('title');
      $href = $link->getAttribute('href');
      if (!empty($title) || empty($href)) {
        continue;
      }

      $url_title = $this->getTitleFromUrl($href);
      if (!empty($url_title)) {
        $link->setAttribute('title', $url_title);
      }
    }

    $result->setProcessedText($dom_document->saveHTML());
    return $result;
  }

  /**
   * {@inheritdoc}
   */
  public function tips($long = FALSE) {
    return $this->t('Adds a title attribute to links found in the content.');
  }

  /**
   * Helper that retrieves the page title for a remote URL.
   *
   * Use a Guzzle stream to prevent the whole page from being read. The PHP
   * stream API supports reading a stream to a given string through the
   * `stream_get_line`, however that isn't implemented in the Guzzle stream.
   * This method reads the stream in manageble chunks untill the `</title>` or
   * `<body>` tag is found.
   *
   * @param string $url
   *   The URL to retrieve the page title from.
   *
   * @return string
   *   The page title or an empty string.
   */
  private function getTitleFromUrl(string $url) {
    // Add support for internal urls.
    if (!UrlHelper::isExternal($url)) {
      $url = Url::fromUserInput($url, [
        'absolute' => TRUE,
      ])->toString();
    }

    try {
      $response = $this->client->request('GET', $url, [
        RequestOptions::STREAM => TRUE,
        RequestOptions::ALLOW_REDIRECTS => FALSE,
        RequestOptions::CONNECT_TIMEOUT => 5,
      ]);
    }
    catch (GuzzleException $e) {
      watchdog_exception('linktitle', $e);
      return '';
    }

    $buffer = '';
    while (!$response->getBody()->eof()) {
      $buffer .= $response->getBody()->read(self::SEEK_LENGTH);
      if (FALSE !== ($title_end = strpos($buffer, '</title>'))) {
        $response->getBody()->close();
        break;
      }

      if (FALSE !== strpos($buffer, '<body>')) {
        $response->getBody()->close();
        return '';
      }
    }

    if (FALSE === ($title_start = strpos($buffer, '<title>'))) {
      return '';
    }

    // Account for the tags.
    return substr($buffer, ($title_start + 7), ($title_end - $title_start - 7));
  }

}
